# Copyright 2016 Huawei.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import portbindings
from neutron_lib import context

from oslo_utils import importutils

from neutron.api import extensions as api_ext
from neutron.common import config

from networking_sfc.db import flowclassifier_db as fdb
from networking_sfc.extensions import flowclassifier
from networking_sfc.services.flowclassifier.common import context as fc_ctx
from networking_sfc.services.flowclassifier.common import exceptions as fc_exc
from networking_sfc.services.flowclassifier.drivers.ovs import driver
from networking_sfc.tests import base
from networking_sfc.tests.unit.db import test_flowclassifier_db


class OVSFlowClassifierDriverTestCase(
    test_flowclassifier_db.FlowClassifierDbPluginTestCaseBase,
    base.NeutronDbPluginV2TestCase
):

    resource_prefix_map = dict([
        (k, flowclassifier.FLOW_CLASSIFIER_PREFIX)
        for k in flowclassifier.RESOURCE_ATTRIBUTE_MAP.keys()
    ])

    def setUp(self):
        flowclassifier_plugin = (
            test_flowclassifier_db.DB_FLOWCLASSIFIER_PLUGIN_CLASS)

        service_plugins = {
            flowclassifier.FLOW_CLASSIFIER_EXT: flowclassifier_plugin
        }
        fdb.FlowClassifierDbPlugin.supported_extension_aliases = [
            flowclassifier.FLOW_CLASSIFIER_EXT]
        fdb.FlowClassifierDbPlugin.path_prefix = (
            flowclassifier.FLOW_CLASSIFIER_PREFIX
        )
        super(OVSFlowClassifierDriverTestCase, self).setUp(
            ext_mgr=None,
            plugin=None,
            service_plugins=service_plugins
        )
        self.flowclassifier_plugin = importutils.import_object(
            flowclassifier_plugin)
        ext_mgr = api_ext.PluginAwareExtensionManager(
            test_flowclassifier_db.extensions_path,
            {
                flowclassifier.FLOW_CLASSIFIER_EXT: self.flowclassifier_plugin
            }
        )
        app = config.load_paste_app('extensions_test_app')
        self.ext_api = api_ext.ExtensionMiddleware(app, ext_mgr=ext_mgr)
        self.ctx = context.get_admin_context()
        self.driver = driver.OVSFlowClassifierDriver()
        self.driver.initialize()

    def tearDown(self):
        super(OVSFlowClassifierDriverTestCase, self).tearDown()

    def test_create_flow_classifier_precommit(self):
        with self.port(
            name='port1',
            device_owner='compute',
            device_id='test',
            arg_list=(
                portbindings.HOST_ID,
            ),
            **{portbindings.HOST_ID: 'test'}
        ) as src_port:
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'logical_source_port': src_port['port']['id']
            }) as fc:
                fc_context = fc_ctx.FlowClassifierContext(
                    self.flowclassifier_plugin, self.ctx,
                    fc['flow_classifier']
                )
                self.driver.create_flow_classifier_precommit(fc_context)

    def test_create_flow_classifier_precommit_no_logical_source_port(self):
        with self.flow_classifier(flow_classifier={
            'name': 'test1',
            'logical_source_port': None
        }) as fc:
            fc_context = fc_ctx.FlowClassifierContext(
                self.flowclassifier_plugin, self.ctx,
                fc['flow_classifier']
            )
            self.assertRaises(
                fc_exc.FlowClassifierBadRequest,
                self.driver.create_flow_classifier_precommit,
                fc_context)



# Copyright 2015 Futurewei.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Initial Mitaka no-op script.

Revision ID: 24fc7241aa5
Revises: start_networking_sfc
Create Date: 2015-09-11 11:37:19.349951

"""

from neutron.db.migration import cli


# revision identifiers, used by Alembic.
revision = '24fc7241aa5'
down_revision = 'start_networking_sfc'
branch_labels = (cli.EXPAND_BRANCH,)


def upgrade():
    pass



# Copyright 2015 Futurewei.  All rights reserved.
# Copyright 2017 Intel Corporation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import uuidutils

import sqlalchemy as sa
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy import orm
from sqlalchemy.orm import backref
from sqlalchemy.orm.collections import attribute_mapped_collection
from sqlalchemy.orm import exc

from neutron.db import models_v2
from neutron_lib.db import api as db_api
from neutron_lib.db import constants as db_const
from neutron_lib.db import model_base
from neutron_lib.db import model_query
from neutron_lib.db import utils as db_utils

from networking_sfc.db import flowclassifier_db as fc_db
from networking_sfc.extensions import flowclassifier as ext_fc
from networking_sfc.extensions import servicegraph as ext_sg
from networking_sfc.extensions import sfc as ext_sfc
from networking_sfc.extensions import tap as ext_tap


LOG = logging.getLogger(__name__)

UUID_LEN = 36
PARAM_LEN = 255
VAR_MAX_LEN = 1024


class ChainParameter(model_base.BASEV2):
    """Represents a single chain parameter."""
    __tablename__ = 'sfc_port_chain_parameters'
    keyword = sa.Column(sa.String(PARAM_LEN), primary_key=True)
    value = sa.Column(sa.String(PARAM_LEN))
    chain_id = sa.Column(
        sa.String(UUID_LEN),
        sa.ForeignKey('sfc_port_chains.id', ondelete='CASCADE'),
        primary_key=True)


class ServiceFunctionParam(model_base.BASEV2):
    """Represents a service function parameter."""
    __tablename__ = 'sfc_service_function_params'
    keyword = sa.Column(sa.String(PARAM_LEN), primary_key=True)
    value = sa.Column(sa.String(PARAM_LEN))
    pair_id = sa.Column(
        sa.String(UUID_LEN),
        sa.ForeignKey('sfc_port_pairs.id', ondelete='CASCADE'),
        primary_key=True)


class PortPairGroupParam(model_base.BASEV2):
    """Represents a port pair group parameter."""
    __tablename__ = 'sfc_port_pair_group_params'
    keyword = sa.Column(sa.String(PARAM_LEN), primary_key=True)
    value = sa.Column(sa.String(VAR_MAX_LEN))
    pair_group_id = sa.Column(
        sa.String(UUID_LEN),
        sa.ForeignKey('sfc_port_pair_groups.id', ondelete='CASCADE'),
        primary_key=True)


class ChainClassifierAssoc(model_base.BASEV2):
    """Relation table between sfc_port_chains and flow_classifiers."""
    __tablename__ = 'sfc_chain_classifier_associations'
    flowclassifier_id = sa.Column(
        sa.String(UUID_LEN),
        sa.ForeignKey('sfc_flow_classifiers.id', ondelete='RESTRICT'),
        primary_key=True, nullable=False, unique=True)
    portchain_id = sa.Column(
        sa.String(UUID_LEN),
        sa.ForeignKey('sfc_port_chains.id', ondelete='CASCADE'),
        primary_key=True)
    flow_classifier = orm.relationship(
        fc_db.FlowClassifier,
        backref=backref('chain_classifier_association', uselist=False),
        uselist=False
    )


class PortPair(model_base.BASEV2, model_base.HasId, model_base.HasProject):
    """Represents the ingress and egress ports for a single service function.

    """
    __tablename__ = 'sfc_port_pairs'
    name = sa.Column(sa.String(db_const.NAME_FIELD_SIZE))
    description = sa.Column(sa.String(db_const.DESCRIPTION_FIELD_SIZE))
    ingress = sa.Column(
        sa.String(UUID_LEN),
        sa.ForeignKey('ports.id', ondelete='RESTRICT'),
        nullable=False)
    egress = sa.Column(
        sa.String(UUID_LEN),
        sa.ForeignKey('ports.id', ondelete='RESTRICT'),
        nullable=False)

    portpairgroup_id = sa.Column(
        sa.String(UUID_LEN),
        sa.ForeignKey('sfc_port_pair_groups.id', ondelete='RESTRICT'))
    service_function_parameters = orm.relationship(
        ServiceFunctionParam,
        collection_class=attribute_mapped_collection('keyword'),
        cascade='all, delete-orphan')

    __table_args__ = (
        sa.UniqueConstraint(
            ingress, egress,
            name='uniq_sfc_port_pairs0ingress0egress'
        ),
        model_base.BASEV2.__table_args__
    )


class ChainGroupAssoc(model_base.BASEV2):
    """Relation table between sfc_port_chains and sfc_port_pair_groups."""
    __tablename__ = 'sfc_chain_group_associations'
    portpairgroup_id = sa.Column(
        sa.String(UUID_LEN),
        sa.ForeignKey('sfc_port_pair_groups.id', ondelete='RESTRICT'),
        primary_key=True, nullable=False)
    portchain_id = sa.Column(
        sa.String(UUID_LEN),
        sa.ForeignKey('sfc_port_chains.id', ondelete='CASCADE'),
        primary_key=True)
    position = sa.Column(sa.Integer)


class PortPairGroup(model_base.BASEV2, model_base.HasId,
                    model_base.HasProject):
    """Represents a port pair group model."""
    __tablename__ = 'sfc_port_pair_groups'
    group_id = sa.Column(sa.Integer(), unique=True, nullable=False)
    name = sa.Column(sa.String(db_const.NAME_FIELD_SIZE))
    description = sa.Column(sa.String(db_const.DESCRIPTION_FIELD_SIZE))
    tap_enabled = sa.Column(sa.Boolean(), server_default=sa.sql.false(),
                            nullable=False)
    port_pairs = orm.relationship(
        PortPair,
        backref='port_pair_group'
    )
    port_pair_group_parameters = orm.relationship(
        PortPairGroupParam,
        collection_class=attribute_mapped_collection('keyword'),
        cascade='all, delete-orphan')
    chain_group_associations = orm.relationship(
        ChainGroupAssoc,
        backref='port_pair_groups')


class PortChain(model_base.BASEV2, model_base.HasId, model_base.HasProject):
    """Represents a Neutron service function Port Chain."""
    __tablename__ = 'sfc_port_chains'
    chain_id = sa.Column(sa.Integer(), unique=True, nullable=False)
    name = sa.Column(sa.String(db_const.NAME_FIELD_SIZE))
    description = sa.Column(sa.String(db_const.DESCRIPTION_FIELD_SIZE))
    chain_group_associations = orm.relationship(
        ChainGroupAssoc,
        backref='port_chain',
        order_by="ChainGroupAssoc.position",
        collection_class=ordering_list('position'),
        cascade='all, delete-orphan')
    chain_classifier_associations = orm.relationship(
        ChainClassifierAssoc,
        backref='port_chain',
        cascade='all, delete-orphan')
    chain_parameters = orm.relationship(
        ChainParameter,
        collection_class=attribute_mapped_collection('keyword'),
        cascade='all, delete-orphan')


class GraphChainAssoc(model_base.BASEV2):
    """Relation table between Service Graphs and src+dst Port Chains."""
    __tablename__ = 'sfc_service_graph_chain_associations'
    service_graph_id = sa.Column(
        sa.String(UUID_LEN),
        sa.ForeignKey('sfc_service_graphs.id', ondelete='CASCADE'),
        primary_key=True, nullable=False)
    src_chain = sa.Column(
        sa.String(UUID_LEN),
        sa.ForeignKey('sfc_port_chains.id', ondelete='RESTRICT'),
        primary_key=True, nullable=False)
    dst_chain = sa.Column(
        sa.String(UUID_LEN),
        sa.ForeignKey('sfc_port_chains.id', ondelete='RESTRICT'),
        primary_key=True, nullable=False)


class ServiceGraph(model_base.BASEV2, model_base.HasId, model_base.HasProject):
    """Represents a Neutron Service Function Chain Graph."""
    __tablename__ = 'sfc_service_graphs'
    name = sa.Column(sa.String(db_const.NAME_FIELD_SIZE))
    description = sa.Column(sa.String(db_const.DESCRIPTION_FIELD_SIZE))
    graph_chain_associations = orm.relationship(
        GraphChainAssoc,
        backref='service_graph',
        cascade='all, delete-orphan')


class SfcDbPlugin(
    ext_sfc.SfcPluginBase,
    ext_sg.ServiceGraphPluginBase
):
    """Mixin class to add port chain to db_plugin_base_v2."""

    def _make_port_chain_dict(self, port_chain, fields=None):
        res = {
            'id': port_chain['id'],
            'name': port_chain['name'],
            'project_id': port_chain['project_id'],
            'description': port_chain['description'],
            'port_pair_groups': [
                assoc['portpairgroup_id']
                for assoc in port_chain['chain_group_associations']
            ],
            'flow_classifiers': [
                assoc['flowclassifier_id']
                for assoc in port_chain['chain_classifier_associations']
            ],
            'chain_parameters': {
                param['keyword']: jsonutils.loads(param['value'])
                for k, param in port_chain['chain_parameters'].items()
            },
            'chain_id': port_chain['chain_id'],
        }
        return db_utils.resource_fields(res, fields)

    @db_api.CONTEXT_READER
    def _validate_port_pair_groups(self, context, pg_ids, pc_id=None):
        prev_pg_tap_enabled = False
        for pg_id in pg_ids:
            pg = self._get_port_pair_group(context, pg_id)
            curr_pg_tap_enabled = pg['tap_enabled']
            if prev_pg_tap_enabled and curr_pg_tap_enabled:
                raise ext_tap.ConsecutiveTapPPGNotSupported()
            prev_pg_tap_enabled = curr_pg_tap_enabled
        query = model_query.query_with_hooks(context, PortChain)
        for port_chain_db in query.all():
            if port_chain_db['id'] == pc_id:
                continue
            pc_pg_ids = [
                assoc['portpairgroup_id']
                for assoc in port_chain_db.chain_group_associations
            ]
            if pc_pg_ids and pg_ids and pc_pg_ids == pg_ids:
                raise ext_sfc.InvalidPortPairGroups(
                    port_pair_groups=pg_ids, port_chain=port_chain_db.id)

    @db_api.CONTEXT_READER
    def _validate_correlation_consistency(self, context, ppg_ids, pc_corr):
        # format like in ServiceFunctionParam.value to aid comparison later:
        pc_corr = jsonutils.dumps(pc_corr)
        for ppg_id in ppg_ids:
            ppg = self._get_port_pair_group(context, ppg_id)
            for pp in ppg['port_pairs']:
                pp_corr = pp['service_function_parameters']['correlation']
                if pp_corr.value != 'null' and pp_corr.value != pc_corr:
                    raise ext_sfc.PortChainInconsistentCorrelations(
                        ppg=ppg_id)

    @db_api.CONTEXT_READER
    def _validate_flow_classifiers(self, context, fc_ids, pc_id=None):
        fcs = [
            self._get_flow_classifier(context, fc_id)
            for fc_id in fc_ids
        ]
        for fc in fcs:
            fc_assoc = fc.chain_classifier_association
            if fc_assoc and fc_assoc['portchain_id'] != pc_id:
                raise ext_fc.FlowClassifierInUse(id=fc.id)

        query = model_query.query_with_hooks(context, PortChain)
        for port_chain_db in query.all():
            if port_chain_db['id'] == pc_id:
                continue
            pc_fc_ids = [
                assoc['flowclassifier_id']
                for assoc in port_chain_db.chain_classifier_associations
            ]
            pc_fcs = [
                self._get_flow_classifier(context, pc_fc_id)
                for pc_fc_id in pc_fc_ids
            ]
            for pc_fc in pc_fcs:
                for fc in fcs:
                    fc_cls = fc_db.FlowClassifierDbPlugin
                    if fc_cls.flowclassifier_basic_conflict(
                        pc_fc, fc
                    ):
                        raise ext_sfc.PortChainFlowClassifierInConflict(
                            fc_id=fc['id'], pc_id=port_chain_db['id'],
                            pc_fc_id=pc_fc['id']
                        )

    @db_api.CONTEXT_READER
    def _setup_chain_group_associations(
        self, context, port_chain, pg_ids
    ):
        chain_group_associations = []
        for pg_id in pg_ids:
            query = model_query.query_with_hooks(context, ChainGroupAssoc)
            chain_group_association = query.filter_by(
                portchain_id=port_chain.id, portpairgroup_id=pg_id
            ).first()
            if not chain_group_association:
                chain_group_association = ChainGroupAssoc(
                    portpairgroup_id=pg_id
                )
            chain_group_associations.append(chain_group_association)
        port_chain.chain_group_associations = chain_group_associations

    @db_api.CONTEXT_READER
    def _setup_chain_classifier_associations(
        self, context, port_chain, fc_ids
    ):
        chain_classifier_associations = []
        for fc_id in fc_ids:
            query = model_query.query_with_hooks(
                context, ChainClassifierAssoc)
            chain_classifier_association = query.filter_by(
                portchain_id=port_chain.id, flowclassifier_id=fc_id
            ).first()
            if not chain_classifier_association:
                chain_classifier_association = ChainClassifierAssoc(
                    flowclassifier_id=fc_id
                )
            chain_classifier_associations.append(
                chain_classifier_association)
        port_chain.chain_classifier_associations = (
            chain_classifier_associations)

    @log_helpers.log_method_call
    def create_port_chain(self, context, port_chain):
        """Create a port chain."""
        pc = port_chain['port_chain']
        project_id = pc['project_id']
        chain_id = pc['chain_id']
        with db_api.CONTEXT_WRITER.using(context):
            chain_parameters = {
                key: ChainParameter(keyword=key, value=jsonutils.dumps(val))
                for key, val in pc['chain_parameters'].items()}

            ppg_ids = pc['port_pair_groups']
            fc_ids = pc['flow_classifiers']
            self._validate_port_pair_groups(context, ppg_ids)
            self._validate_correlation_consistency(
                context, ppg_ids, pc['chain_parameters']['correlation'])
            self._validate_flow_classifiers(context, fc_ids)
            assigned_chain_ids = {}
            query = context.session.query(PortChain)
            for port_chain_db in query.all():
                assigned_chain_ids[port_chain_db['chain_id']] = (
                    port_chain_db['id']
                )
            if not chain_id:
                available_chain_id = 1
                while available_chain_id < ext_sfc.MAX_CHAIN_ID:
                    if available_chain_id not in assigned_chain_ids:
                        chain_id = available_chain_id
                        break
                    available_chain_id += 1
                if not chain_id:
                    raise ext_sfc.PortChainUnavailableChainId()
            else:
                if chain_id in assigned_chain_ids:
                    raise ext_sfc.PortChainChainIdInConflict(
                        chain_id=chain_id, pc_id=assigned_chain_ids[chain_id])
            port_chain_db = PortChain(id=uuidutils.generate_uuid(),
                                      project_id=project_id,
                                      description=pc['description'],
                                      name=pc['name'],
                                      chain_parameters=chain_parameters,
                                      chain_id=chain_id)
            self._setup_chain_group_associations(
                context, port_chain_db, ppg_ids)
            self._setup_chain_classifier_associations(
                context, port_chain_db, fc_ids)
            context.session.add(port_chain_db)

            return self._make_port_chain_dict(port_chain_db)

    @log_helpers.log_method_call
    @db_api.CONTEXT_READER
    def get_port_chains(self, context, filters=None, fields=None,
                        sorts=None, limit=None,
                        marker=None, page_reverse=False, default_sg=False):
        marker_obj = db_utils.get_marker_obj(self, context, 'port_chain',
                                             limit, marker)
        return model_query.get_collection(
            context, PortChain,
            self._make_port_chain_dict,
            filters=filters, fields=fields,
            sorts=sorts,
            limit=limit, marker_obj=marker_obj,
            page_reverse=page_reverse)

    @db_api.CONTEXT_READER
    def get_port_chains_count(self, context, filters=None):
        return model_query.get_collection_count(
            context, PortChain, filters=filters)

    @log_helpers.log_method_call
    @db_api.CONTEXT_READER
    def get_port_chain(self, context, id, fields=None):
        portchain = self._get_port_chain(context, id)
        return self._make_port_chain_dict(portchain, fields)

    @log_helpers.log_method_call
    def _get_port_chain(self, context, id):
        try:
            return model_query.get_by_id(context, PortChain, id)
        except exc.NoResultFound:
            raise ext_sfc.PortChainNotFound(id=id)

    @log_helpers.log_method_call
    def delete_port_chain(self, context, id):
        # if this port chain is part of a graph, abort delete:
        if self._any_port_chains_in_a_graph(context, {id}):
            raise ext_sg.ServiceGraphPortChainInUse(id=id)
        try:
            with db_api.CONTEXT_WRITER.using(context):
                pc = self._get_port_chain(context, id)
                context.session.delete(pc)
        except ext_sfc.PortChainNotFound:
            LOG.info("Deleting a non-existing port chain.")

    @log_helpers.log_method_call
    def update_port_chain(self, context, id, port_chain):
        pc = port_chain['port_chain']
        # if this port chain is part of a graph, abort non-neutral updates:
        if self._any_port_chains_in_a_graph(context, {id}):
            if 'flow_classifiers' in pc or 'port_pair_groups' in pc:
                raise ext_sg.ServiceGraphPortChainInUse(id=id)

        with db_api.CONTEXT_WRITER.using(context):
            pc_db = self._get_port_chain(context, id)
            for k, v in pc.items():
                if k == 'flow_classifiers':
                    self._validate_flow_classifiers(
                        context, v, pc_id=id)
                    self._setup_chain_classifier_associations(
                        context, pc_db, v)
                elif k == 'port_pair_groups':
                    self._validate_port_pair_groups(
                        context, v, pc_id=id)
                    self._setup_chain_group_associations(
                        context, pc_db, v)
                else:
                    pc_db[k] = v
            return self._make_port_chain_dict(pc_db)

    def _make_port_pair_dict(self, port_pair, fields=None):
        res = {
            'id': port_pair['id'],
            'name': port_pair['name'],
            'description': port_pair['description'],
            'project_id': port_pair['project_id'],
            'ingress': port_pair['ingress'],
            'egress': port_pair['egress'],
            'service_function_parameters': {
                param['keyword']: jsonutils.loads(param['value'])
                for k, param in
                port_pair['service_function_parameters'].items()
            }
        }

        return db_utils.resource_fields(res, fields)

    def _validate_port_pair_ingress_egress(self, ingress, egress):
        if 'device_id' not in ingress or not ingress['device_id']:
            raise ext_sfc.PortPairIngressNoHost(
                ingress=ingress['id']
            )
        if 'device_id' not in egress or not egress['device_id']:
            raise ext_sfc.PortPairEgressNoHost(
                egress=egress['id']
            )
        if ingress['device_id'] != egress['device_id']:
            raise ext_sfc.PortPairIngressEgressDifferentHost(
                ingress=ingress['id'],
                egress=egress['id'])

    @log_helpers.log_method_call
    def create_port_pair(self, context, port_pair):
        """Create a port pair."""
        pp = port_pair['port_pair']
        project_id = pp['project_id']
        with db_api.CONTEXT_WRITER.using(context):
            query = model_query.query_with_hooks(context, PortPair)
            pp_in_use = query.filter_by(
                ingress=pp['ingress'], egress=pp['egress']
            ).first()
            if pp_in_use:
                raise ext_sfc.PortPairIngressEgressInUse(
                    ingress=pp['ingress'],
                    egress=pp['egress'],
                    id=pp_in_use['id']
                )

            service_function_parameters = {
                key: ServiceFunctionParam(
                    keyword=key, value=jsonutils.dumps(val))
                for key, val in
                pp['service_function_parameters'].items()
            }
            ingress = self._get_port(context, pp['ingress'])
            egress = self._get_port(context, pp['egress'])
            self._validate_port_pair_ingress_egress(ingress, egress)
            port_pair_db = PortPair(
                id=uuidutils.generate_uuid(),
                name=pp['name'],
                description=pp['description'],
                project_id=project_id,
                ingress=pp['ingress'],
                egress=pp['egress'],
                service_function_parameters=service_function_parameters
            )
            context.session.add(port_pair_db)
            return self._make_port_pair_dict(port_pair_db)

    @log_helpers.log_method_call
    @db_api.CONTEXT_READER
    def get_port_pairs(self, context, filters=None, fields=None,
                       sorts=None, limit=None, marker=None,
                       page_reverse=False):
        marker_obj = db_utils.get_marker_obj(self, context, 'port_pair',
                                             limit, marker)
        return model_query.get_collection(
            context, PortPair,
            self._make_port_pair_dict,
            filters=filters, fields=fields,
            sorts=sorts,
            limit=limit, marker_obj=marker_obj,
            page_reverse=page_reverse)

    @db_api.CONTEXT_READER
    def get_port_pairs_count(self, context, filters=None):
        return model_query.get_collection_count(
            context, PortPair, filters=filters)

    @log_helpers.log_method_call
    @db_api.CONTEXT_READER
    def get_port_pair(self, context, id, fields=None):
        port_pair = self._get_port_pair(context, id)
        return self._make_port_pair_dict(port_pair, fields)

    def _get_port_pair(self, context, id):
        try:
            return model_query.get_by_id(context, PortPair, id)
        except exc.NoResultFound:
            raise ext_sfc.PortPairNotFound(id=id)

    def _get_port(self, context, id):
        try:
            return model_query.get_by_id(context, models_v2.Port, id)
        except exc.NoResultFound:
            raise ext_sfc.PortPairPortNotFound(id=id)

    @log_helpers.log_method_call
    def update_port_pair(self, context, id, port_pair):
        new_pp = port_pair['port_pair']
        with db_api.CONTEXT_WRITER.using(context):
            old_pp = self._get_port_pair(context, id)
            old_pp.update(new_pp)
            return self._make_port_pair_dict(old_pp)

    @log_helpers.log_method_call
    def delete_port_pair(self, context, id):
        try:
            with db_api.CONTEXT_WRITER.using(context):
                pp = self._get_port_pair(context, id)
                if pp.portpairgroup_id:
                    raise ext_sfc.PortPairInUse(id=id)
                context.session.delete(pp)
        except ext_sfc.PortPairNotFound:
            LOG.info("Deleting a non-existing port pair.")

    def _make_port_pair_group_dict(self, port_pair_group, fields=None):
        res = {
            'id': port_pair_group['id'],
            'name': port_pair_group['name'],
            'description': port_pair_group['description'],
            'project_id': port_pair_group['project_id'],
            'port_pairs': [pp['id'] for pp in port_pair_group['port_pairs']],
            'tap_enabled': port_pair_group['tap_enabled'],
            'port_pair_group_parameters': {
                param['keyword']: jsonutils.loads(param['value'])
                for k, param in
                port_pair_group['port_pair_group_parameters'].items()
            },
            'group_id': port_pair_group.get('group_id') or 0
        }

        return db_utils.resource_fields(res, fields)

    def _validate_pps_in_ppg(self, portpairs_list, id=None):
        first_check = True
        correlation = None
        for portpair in portpairs_list:
            sfparams = portpair.service_function_parameters
            pp_corr = sfparams['correlation']
            if first_check:
                first_check = False
                correlation = pp_corr.value
            if pp_corr.value != correlation:
                # don't include PPs of different correlations
                raise ext_sfc.InconsistentCorrelations()
            if (
                portpair.portpairgroup_id and
                portpair.portpairgroup_id != id
            ):
                # don't include PPs included by other PPGs
                raise ext_sfc.PortPairInUse(id=portpair.id)

    @log_helpers.log_method_call
    def create_port_pair_group(self, context, port_pair_group):
        """Create a port pair group."""
        pg = port_pair_group['port_pair_group']
        project_id = pg['project_id']

        with db_api.CONTEXT_WRITER.using(context):
            portpairs_list = [self._get_port_pair(context, pp_id)
                              for pp_id in pg['port_pairs']]
            self._validate_pps_in_ppg(portpairs_list)
            if pg['tap_enabled'] and len(portpairs_list) > 1:
                raise ext_tap.MultiplePortPairsInTapPPGNotSupported()
            port_pair_group_parameters = {
                key: PortPairGroupParam(
                    keyword=key, value=jsonutils.dumps(val))
                for key, val in
                pg['port_pair_group_parameters'].items()
            }
            assigned_group_ids = {}
            query = context.session.query(PortPairGroup)
            for port_pair_group_db in query.all():
                assigned_group_ids[port_pair_group_db['group_id']] = (
                    port_pair_group_db['id']
                )
            group_id = 0
            available_group_id = 1
            while True:
                if available_group_id not in assigned_group_ids:
                    group_id = available_group_id
                    break
                available_group_id += 1
            port_pair_group_db = PortPairGroup(
                id=uuidutils.generate_uuid(),
                name=pg['name'],
                description=pg['description'],
                project_id=project_id,
                port_pairs=portpairs_list,
                port_pair_group_parameters=port_pair_group_parameters,
                group_id=group_id,
                tap_enabled=pg['tap_enabled']
            )
            context.session.add(port_pair_group_db)
            return self._make_port_pair_group_dict(port_pair_group_db)

    @log_helpers.log_method_call
    @db_api.CONTEXT_READER
    def get_port_pair_groups(self, context, filters=None, fields=None,
                             sorts=None, limit=None, marker=None,
                             page_reverse=False):
        marker_obj = db_utils.get_marker_obj(self, context, 'port_pair_group',
                                             limit, marker)
        return model_query.get_collection(
            context, PortPairGroup,
            self._make_port_pair_group_dict,
            filters=filters, fields=fields,
            sorts=sorts,
            limit=limit, marker_obj=marker_obj,
            page_reverse=page_reverse)

    @db_api.CONTEXT_READER
    def get_port_pair_groups_count(self, context, filters=None):
        return model_query.get_collection_count(
            context, PortPairGroup, filters=filters)

    @log_helpers.log_method_call
    @db_api.CONTEXT_READER
    def get_port_pair_group(self, context, id, fields=None):
        port_pair_group = self._get_port_pair_group(context, id)
        return self._make_port_pair_group_dict(port_pair_group, fields)

    def _get_port_pair_group(self, context, id):
        try:
            return model_query.get_by_id(context, PortPairGroup, id)
        except exc.NoResultFound:
            raise ext_sfc.PortPairGroupNotFound(id=id)

    @db_api.CONTEXT_READER
    def _get_flow_classifier(self, context, id):
        try:
            return model_query.get_by_id(context, fc_db.FlowClassifier, id)
        except exc.NoResultFound:
            raise ext_fc.FlowClassifierNotFound(id=id)

    @log_helpers.log_method_call
    def update_port_pair_group(self, context, id, port_pair_group):
        new_pg = port_pair_group['port_pair_group']

        with db_api.CONTEXT_WRITER.using(context):
            portpairs_list = [self._get_port_pair(context, pp_id)
                              for pp_id in new_pg.get('port_pairs', [])]
            self._validate_pps_in_ppg(portpairs_list, id)
            old_pg = self._get_port_pair_group(context, id)
            if old_pg['tap_enabled'] and len(portpairs_list) > 1:
                raise ext_tap.MultiplePortPairsInTapPPGNotSupported()
            for k, v in new_pg.items():
                if k == 'port_pairs':
                    port_pairs = [
                        self._get_port_pair(context, pp_id)
                        for pp_id in v
                    ]
                    old_pg.port_pairs = port_pairs
                else:
                    old_pg[k] = v

            return self._make_port_pair_group_dict(old_pg)

    @log_helpers.log_method_call
    def delete_port_pair_group(self, context, id):
        try:
            with db_api.CONTEXT_WRITER.using(context):
                pg = self._get_port_pair_group(context, id)
                if pg.chain_group_associations:
                    raise ext_sfc.PortPairGroupInUse(id=id)
                context.session.delete(pg)
        except ext_sfc.PortPairGroupNotFound:
            LOG.info("Deleting a non-existing port pair group.")

    def _graph_assocs_to_pc_dict(self, assocs):
        assoc_dict = {}
        for assoc in assocs:
            if assoc.src_chain in assoc_dict:
                assoc_dict[assoc.src_chain].append(assoc.dst_chain)
            else:
                assoc_dict[assoc.src_chain] = [assoc.dst_chain]
        return assoc_dict

    def _make_service_graph_dict(self, graph_db, fields=None):
        res = {
            'id': graph_db['id'],
            'name': graph_db['name'],
            'project_id': graph_db['project_id'],
            'description': graph_db['description'],
            'port_chains': self._graph_assocs_to_pc_dict(
                graph_db['graph_chain_associations'])
        }
        return db_utils.resource_fields(res, fields)

    def _make_graph_chain_assoc_dict(self, assoc_db, fields=None):
        res = {
            'service_graph_id': assoc_db['service_graph_id'],
            'src_chain': assoc_db['src_chain'],
            'dst_chain': assoc_db['dst_chain']
        }
        return db_utils.resource_fields(res, fields)

    def _is_there_a_loop(self, parenthood, current_chain, traversed):
        if current_chain not in parenthood:
            return False
        if current_chain not in traversed:
            loop_status = False
            traversed.append(current_chain)
            for port_chain in parenthood[current_chain]:
                loop_status = loop_status or self._is_there_a_loop(
                    parenthood, port_chain, list(traversed))
            return loop_status
        return True

    def _any_port_chains_in_a_graph(self, context,
                                    port_chains=set(), graph_id=None):
        if not port_chains:
            return False
        with db_api.CONTEXT_READER.using(context):
            query = model_query.query_with_hooks(context, ServiceGraph)
            for graph_db in query.all():
                if graph_db['id'] == graph_id:
                    continue
                pc_ids = [
                    assoc['src_chain']
                    for assoc in graph_db.graph_chain_associations
                ]
                pc_ids.extend([
                    assoc['dst_chain']
                    for assoc in graph_db.graph_chain_associations
                ])
                if pc_ids and port_chains and set(
                        pc_ids).intersection(port_chains):
                    return True
        return False

    def _validate_port_chains_for_graph(self, context,
                                        port_chains, graph_id=None):
        # create a list of all port-chains that will be associated
        all_port_chains = set()
        for src_chain in port_chains:
            all_port_chains.add(src_chain)
            for dst_chain in port_chains[src_chain]:
                all_port_chains.add(dst_chain)
        # check if any of the port-chains are already in a graph
        if self._any_port_chains_in_a_graph(
                context, all_port_chains, graph_id):
            raise ext_sg.ServiceGraphInvalidPortChains(
                port_chains=port_chains)

        # dict whose keys are PCs and values are lists of dependency-PCs
        # (PCs incoming to the point where the key is a outgoing)
        parenthood = {}
        encapsulation = None
        fc_cls = fc_db.FlowClassifierDbPlugin
        for src_chain in port_chains:
            src_pc = self._get_port_chain(context, src_chain)
            curr_corr = src_pc.chain_parameters['correlation']['value']
            # guarantee that branching PPG supports correlation
            assocs = src_pc.chain_group_associations
            src_ppg = max(assocs, key=(lambda ppg: ppg.position))
            ppg_id = src_ppg['portpairgroup_id']
            ppg = self._get_port_pair_group(context, ppg_id)
            for pp in ppg.port_pairs:
                sfparams = pp['service_function_parameters']
                if sfparams['correlation']['value'] != curr_corr:
                    raise ext_sg.ServiceGraphImpossibleBranching()

            # verify encapsulation consistency across all PCs (part 1)
            if not encapsulation:
                encapsulation = curr_corr
            elif encapsulation != curr_corr:
                raise ext_sg.ServiceGraphInconsistentEncapsulation()
            # list of all port chains at this branching point:
            branching_point = []
            # list of every flow classifier at this branching point:
            fcs_for_src_chain = []
            for dst_chain in port_chains[src_chain]:
                # check if the current destination PC was already added
                if dst_chain in branching_point:
                    raise ext_sg.ServiceGraphPortChainInConflict(
                        pc_id=dst_chain)
                branching_point.append(dst_chain)
                dst_pc = self._get_port_chain(context, dst_chain)
                curr_corr = dst_pc.chain_parameters['correlation']['value']
                # guarantee that destination PPG supports correlation
                assocs = dst_pc.chain_group_associations
                dst_ppg = min(assocs, key=(lambda ppg: ppg.position))
                ppg_id = dst_ppg['portpairgroup_id']
                ppg = self._get_port_pair_group(context, ppg_id)
                for pp in ppg.port_pairs:
                    sfparams = pp['service_function_parameters']
                    if sfparams['correlation']['value'] != curr_corr:
                        raise ext_sg.ServiceGraphImpossibleBranching()
                # verify encapsulation consistency across all PCs (part 2)
                if encapsulation != curr_corr:
                    raise ext_sg.ServiceGraphInconsistentEncapsulation()
                dst_pc_dict = self._make_port_chain_dict(dst_pc)
                # acquire associated flow classifiers
                fcs = dst_pc_dict['flow_classifiers']
                for fc_id in fcs:
                    fc = self._get_flow_classifier(context, fc_id)
                    fcs_for_src_chain.append(fc)  # update list of every FC
                # update the parenthood dict
                if dst_chain in parenthood:
                    parenthood[dst_chain].append(src_chain)
                else:
                    parenthood[dst_chain] = [src_chain]
                # detect duplicate FCs, consequently branching ambiguity
                for i, fc1 in enumerate(fcs_for_src_chain):
                    for fc2 in fcs_for_src_chain[i + 1:]:
                        if(fc_cls.flowclassifier_basic_conflict(fc1, fc2)):
                            raise ext_sg.\
                                ServiceGraphFlowClassifierInConflict(
                                    fc1_id=fc1['id'], fc2_id=fc2['id'])

        # check for circular paths within the graph via parenthood dict:
        for port_chain in parenthood:
            if self._is_there_a_loop(parenthood, port_chain, []):
                raise ext_sg.ServiceGraphLoopDetected()

    def _setup_graph_chain_associations(self, context, graph_db, port_chains):
        with db_api.CONTEXT_READER.using(context):
            graph_chain_associations = []
            for src_chain in port_chains:
                query = model_query.query_with_hooks(context, GraphChainAssoc)
                for dst_chain in port_chains[src_chain]:
                    graph_chain_association = query.filter_by(
                        service_graph_id=graph_db.id,
                        src_chain=src_chain, dst_chain=dst_chain).first()
                    if not graph_chain_association:
                        graph_chain_association = GraphChainAssoc(
                            service_graph_id=graph_db.id,
                            src_chain=src_chain,
                            dst_chain=dst_chain
                        )
                    graph_chain_associations.append(graph_chain_association)
            graph_db.graph_chain_associations = graph_chain_associations

    @db_api.CONTEXT_READER
    def get_branches(self, context, filters):
        return model_query.get_collection(
            context, GraphChainAssoc,
            self._make_graph_chain_assoc_dict,
            filters=filters)

    @log_helpers.log_method_call
    def create_service_graph(self, context, service_graph):
        """Create a Service Graph."""
        service_graph = service_graph['service_graph']
        project_id = service_graph['project_id']
        with db_api.CONTEXT_WRITER.using(context):
            port_chains = service_graph['port_chains']
            self._validate_port_chains_for_graph(context, port_chains)
            graph_db = ServiceGraph(id=uuidutils.generate_uuid(),
                                    project_id=project_id,
                                    description=service_graph['description'],
                                    name=service_graph['name'])
            self._setup_graph_chain_associations(
                context, graph_db, port_chains)
            context.session.add(graph_db)
            return self._make_service_graph_dict(graph_db)

    @log_helpers.log_method_call
    @db_api.CONTEXT_READER
    def get_service_graphs(self, context, filters=None,
                           fields=None, sorts=None, limit=None,
                           marker=None, page_reverse=False):
        """Get Service Graphs."""
        marker_obj = db_utils.get_marker_obj(self, context,
                                             'service_graph', limit, marker)
        return model_query.get_collection(
            context, ServiceGraph,
            self._make_service_graph_dict,
            filters=filters, fields=fields,
            sorts=sorts,
            limit=limit, marker_obj=marker_obj,
            page_reverse=page_reverse)

    @log_helpers.log_method_call
    @db_api.CONTEXT_READER
    def get_service_graph(self, context, id, fields=None):
        """Get a Service Graph."""
        service_graph = self._get_service_graph(context, id)
        return self._make_service_graph_dict(service_graph, fields)

    @log_helpers.log_method_call
    def _get_service_graph(self, context, id):
        try:
            return model_query.get_by_id(context, ServiceGraph, id)
        except exc.NoResultFound:
            raise ext_sg.ServiceGraphNotFound(id=id)

    @log_helpers.log_method_call
    def update_service_graph(self, context, id, service_graph):
        """Update a Service Graph."""
        service_graph = service_graph['service_graph']
        with db_api.CONTEXT_WRITER.using(context):
            graph_db = self._get_service_graph(context, id)
            for k, v in service_graph.items():
                if k == 'port_chains':
                    self._validate_port_chains_for_graph(
                        context, v, graph_id=id)
                    self._setup_graph_chain_associations(
                        context, graph_db, v)
                else:
                    graph_db[k] = v
            return self._make_service_graph_dict(graph_db)

    @log_helpers.log_method_call
    def delete_service_graph(self, context, id):
        """Delete a Service Graph."""
        try:
            with db_api.CONTEXT_WRITER.using(context):
                graph = self._get_service_graph(context, id)
                context.session.delete(graph)
        except ext_sfc.ServiceGraphNotFound:
            LOG.info("Deleting a non-existing Service Graph.")



# Copyright 2015 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class SfcPluginContext():
    """SFC context base class."""
    def __init__(self, plugin, plugin_context):
        self._plugin = plugin
        self._plugin_context = plugin_context


class PortChainContext(SfcPluginContext):

    def __init__(self, plugin, plugin_context, portchain,
                 original_portchain=None):
        super(PortChainContext, self).__init__(plugin, plugin_context)
        self._portchain = portchain
        self._original_portchain = original_portchain

    @property
    def current(self):
        return self._portchain

    @property
    def original(self):
        return self._original_portchain


class FlowClassifierContext(SfcPluginContext):
    def __init__(self, plugin, plugin_context, flowclassifier,
                 original_flowclassifier=None):
        super(FlowClassifierContext, self).__init__(plugin, plugin_context)
        self._flowclassifier = flowclassifier
        self._original_flowclassifier = original_flowclassifier

    @property
    def current(self):
        return self._flowclassifier

    @property
    def original(self):
        return self._original_flowclassifier


class PortPairContext(SfcPluginContext):
    def __init__(self, plugin, plugin_context, portpair,
                 original_portpair=None):
        super(PortPairContext, self).__init__(plugin, plugin_context)
        self._portpair = portpair
        self._original_portpair = original_portpair

    @property
    def current(self):
        return self._portpair

    @property
    def original(self):
        return self._original_portpair


class PortPairGroupContext(SfcPluginContext):
    def __init__(self, plugin, plugin_context, portpairgroup,
                 original_portpairgroup=None):
        super(PortPairGroupContext, self).__init__(plugin, plugin_context)
        self._portpairgroup = portpairgroup
        self._original_portpairgroup = original_portpairgroup

    @property
    def current(self):
        return self._portpairgroup

    @property
    def original(self):
        return self._original_portpairgroup


class ServiceGraphContext(SfcPluginContext):

    def __init__(self, plugin, plugin_context, service_graph,
                 original_graph=None):
        super(ServiceGraphContext, self).__init__(plugin, plugin_context)
        self._service_graph = service_graph
        self._original_graph = original_graph

    @property
    def current(self):
        return self._service_graph

    @property
    def original(self):
        return self._original_graph



# Copyright 2015 Futurewei.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Defining Port Chain data-model.

Revision ID: c3e178d4a985
Revises: 9768e6a66c9
Create Date: 2015-09-11 11:37:19.349951

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'c3e178d4a985'
down_revision = '9768e6a66c9'


def upgrade():
    op.create_table(
        'sfc_port_pair_groups',
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('tenant_id', sa.String(length=255),
                  nullable=True, index=True),
        sa.Column('name', sa.String(length=255),
                  nullable=True),
        sa.Column('description', sa.String(length=255),
                  nullable=True),
        sa.PrimaryKeyConstraint('id')
    )

    op.create_table(
        'sfc_port_pairs',
        sa.Column('tenant_id', sa.String(length=255),
                  nullable=True, index=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('description', sa.String(length=255),
                  nullable=True),
        sa.Column('ingress', sa.String(length=36), nullable=False),
        sa.Column('egress', sa.String(length=36), nullable=False),
        sa.Column('portpairgroup_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['egress'], ['ports.id'],
                                ondelete='RESTRICT'),
        sa.ForeignKeyConstraint(['ingress'], ['ports.id'],
                                ondelete='RESTRICT'),
        sa.ForeignKeyConstraint(['portpairgroup_id'],
                                ['sfc_port_pair_groups.id'],
                                ondelete='RESTRICT'),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('ingress', 'egress',
                            name='uniq_sfc_port_pairs0ingress0egress')
    )

    op.create_table(
        'sfc_port_chains',
        sa.Column('tenant_id', sa.String(length=255),
                  nullable=True, index=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255),
                  nullable=True),
        sa.Column('description', sa.String(length=255),
                  nullable=True),
        sa.PrimaryKeyConstraint('id')
    )

    op.create_table(
        'sfc_chain_group_associations',
        sa.Column('portpairgroup_id', sa.String(length=36), nullable=False),
        sa.Column('portchain_id', sa.String(length=36), nullable=False),
        sa.Column('position', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(['portchain_id'], ['sfc_port_chains.id'], ),
        sa.ForeignKeyConstraint(['portpairgroup_id'],
                                ['sfc_port_pair_groups.id'],
                                ondelete='RESTRICT'),
        sa.PrimaryKeyConstraint('portpairgroup_id', 'portchain_id')
    )

    op.create_table(
        'sfc_port_chain_parameters',
        sa.Column('keyword', sa.String(length=255), nullable=False),
        sa.Column('value', sa.String(length=255), nullable=True),
        sa.Column('chain_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['chain_id'], ['sfc_port_chains.id'], ),
        sa.PrimaryKeyConstraint('keyword', 'chain_id')
    )

    op.create_table(
        'sfc_service_function_params',
        sa.Column('keyword', sa.String(length=255), nullable=False),
        sa.Column('value', sa.String(length=255), nullable=True),
        sa.Column('pair_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['pair_id'], ['sfc_port_pairs.id'], ),
        sa.PrimaryKeyConstraint('keyword', 'pair_id')
    )

    op.create_table(
        'sfc_chain_classifier_associations',
        sa.Column('flowclassifier_id', sa.String(length=36), nullable=False),
        sa.Column('portchain_id', sa.String(length=36), nullable=False),
        sa.ForeignKeyConstraint(['flowclassifier_id'],
                                ['sfc_flow_classifiers.id'],
                                ondelete='RESTRICT'),
        sa.ForeignKeyConstraint(['portchain_id'], ['sfc_port_chains.id'], ),
        sa.PrimaryKeyConstraint('flowclassifier_id', 'portchain_id'),
        sa.UniqueConstraint('flowclassifier_id')
    )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools

from networking_sfc.extensions import flowclassifier
from networking_sfc.extensions import servicegraph
from networking_sfc.extensions import sfc
from networking_sfc.services.flowclassifier.common import config as fc_config
from networking_sfc.services.sfc.common import config as sfc_config


def list_quota_opts():
    return [
        ('quotas',
         itertools.chain(
             flowclassifier.flow_classifier_quota_opts,
             sfc.sfc_quota_opts,
             servicegraph.service_graph_quota_opts)
         ),
    ]


def list_sfc_opts():
    return [
        ('flowclassifier', fc_config.FLOWCLASSIFIER_DRIVER_OPTS),
        ('sfc', sfc_config.SFC_DRIVER_OPTS),
    ]



# Copyright 2015 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from networking_sfc.services.flowclassifier.common import context as fc_ctx
from networking_sfc.services.flowclassifier.common import exceptions as fc_exc
from networking_sfc.tests.unit.db import test_flowclassifier_db

FLOWCLASSIFIER_PLUGIN_KLASS = (
    "networking_sfc.services.flowclassifier."
    "plugin.FlowClassifierPlugin"
)


class FlowClassifierPluginTestCase(
    test_flowclassifier_db.FlowClassifierDbPluginTestCase
):
    def setUp(
        self, core_plugin=None, flowclassifier_plugin=None, ext_mgr=None
    ):
        if not flowclassifier_plugin:
            flowclassifier_plugin = FLOWCLASSIFIER_PLUGIN_KLASS
        self.driver_manager_p = mock.patch(
            'networking_sfc.services.flowclassifier.driver_manager.'
            'FlowClassifierDriverManager'
        )
        self.fake_driver_manager_class = self.driver_manager_p.start()
        self.fake_driver_manager = mock.Mock()
        self.fake_driver_manager_class.return_value = self.fake_driver_manager
        self.plugin_context = None
        self.plugin_context_precommit = None
        self.plugin_context_postcommit = None
        super(FlowClassifierPluginTestCase, self).setUp(
            core_plugin=core_plugin,
            flowclassifier_plugin=flowclassifier_plugin,
            ext_mgr=ext_mgr
        )

    def _record_context(self, plugin_context):
        self.plugin_context = plugin_context

    def _record_context_precommit(self, plugin_context):
        self.plugin_context_precommit = plugin_context

    def _record_context_postcommit(self, plugin_context):
        self.plugin_context_postcommit = plugin_context

    def test_create_flow_classifier_driver_manager_called(self):
        self.fake_driver_manager.create_flow_classifier_precommit = mock.Mock(
            side_effect=self._record_context_precommit)
        self.fake_driver_manager.create_flow_classifier_postcommit = mock.Mock(
            side_effect=self._record_context_postcommit)
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'logical_source_port': port['port']['id']
            }) as fc:
                driver_manager = self.fake_driver_manager
                (driver_manager.create_flow_classifier_precommit
                 .assert_called_once_with(mock.ANY))
                (driver_manager.create_flow_classifier_postcommit
                 .assert_called_once_with(mock.ANY))
                self.assertIsInstance(
                    self.plugin_context_precommit,
                    fc_ctx.FlowClassifierContext)
                self.assertIsInstance(
                    self.plugin_context_postcommit,
                    fc_ctx.FlowClassifierContext)
                self.assertIn('flow_classifier', fc)
                self.assertEqual(
                    self.plugin_context_precommit.current,
                    fc['flow_classifier'])
                self.assertEqual(
                    self.plugin_context_postcommit.current,
                    fc['flow_classifier'])

    def test_create_flow_classifier_postcommit_driver_manager_exception(self):
        self.fake_driver_manager.create_flow_classifier_postcommit = mock.Mock(
            side_effect=fc_exc.FlowClassifierDriverError(
                method='create_flow_classifier_postcommit'
            )
        )
        with self.port(
            name='test1'
        ) as port:
            self._create_flow_classifier(
                self.fmt, {'logical_source_port': port['port']['id']},
                expected_res_status=500)
            driver_manager = self.fake_driver_manager
            (driver_manager.create_flow_classifier_precommit
             .assert_called_once_with(mock.ANY))
            (driver_manager.create_flow_classifier_postcommit
             .assert_called_once_with(mock.ANY))
            (driver_manager.delete_flow_classifier
             .assert_called_once_with(mock.ANY))
            (driver_manager.delete_flow_classifier_precommit
             .assert_called_once_with(mock.ANY))
            (driver_manager.delete_flow_classifier_postcommit
             .assert_called_once_with(mock.ANY))
            self._test_list_resources('flow_classifier', [])

    def test_create_flow_classifier_precommit_driver_manager_exception(self):
        self.fake_driver_manager.create_flow_classifier_precommit = mock.Mock(
            side_effect=fc_exc.FlowClassifierDriverError(
                method='create_flow_classifier_precommit'
            )
        )
        with self.port(
            name='test1'
        ) as port:
            self._test_list_resources('flow_classifier', [])
            self._create_flow_classifier(
                self.fmt, {'logical_source_port': port['port']['id']},
                expected_res_status=500)
            self._test_list_resources('flow_classifier', [])
            driver_manager = self.fake_driver_manager
            (driver_manager.create_flow_classifier_precommit
             .assert_called_once_with(mock.ANY))
            (driver_manager.create_flow_classifier_postcommit
             .assert_not_called())
            driver_manager.delete_flow_classifier.assert_not_called()
            (driver_manager.delete_flow_classifier_precommit
             .assert_not_called())
            (driver_manager.delete_flow_classifier_postcommit
             .assert_not_called())
            self._test_list_resources('flow_classifier', [])

    def test_update_flow_classifier_driver_manager_called(self):
        self.fake_driver_manager.update_flow_classifier_precommit = mock.Mock(
            side_effect=self._record_context_precommit)
        self.fake_driver_manager.update_flow_classifier_postcommit = mock.Mock(
            side_effect=self._record_context_postcommit)
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'logical_source_port': port['port']['id']
            }) as fc:
                req = self.new_update_request(
                    'flow_classifiers', {'flow_classifier': {'name': 'test2'}},
                    fc['flow_classifier']['id']
                )
                res = self.deserialize(
                    self.fmt,
                    req.get_response(self.ext_api)
                )
                driver_manager = self.fake_driver_manager
                (driver_manager.update_flow_classifier_precommit
                 .assert_called_once_with(mock.ANY))
                (driver_manager.update_flow_classifier_postcommit
                 .assert_called_once_with(mock.ANY))
                self.assertIsInstance(
                    self.plugin_context_precommit,
                    fc_ctx.FlowClassifierContext)
                self.assertIsInstance(self.plugin_context_postcommit,
                                      fc_ctx.FlowClassifierContext)
                self.assertIn('flow_classifier', fc)
                self.assertIn('flow_classifier', res)
                self.assertEqual(self.plugin_context_precommit.current,
                                 res['flow_classifier'])
                self.assertEqual(self.plugin_context_postcommit.current,
                                 res['flow_classifier'])
                self.assertEqual(self.plugin_context_precommit.original,
                                 fc['flow_classifier'])
                self.assertEqual(self.plugin_context_postcommit.original,
                                 fc['flow_classifier'])

    def _test_update_flow_classifier_driver_manager_exception(self, updated):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'logical_source_port': port['port']['id']
            }) as fc:
                self.assertIn('flow_classifier', fc)
                original_flow_classifier = fc['flow_classifier']
                req = self.new_update_request(
                    'flow_classifiers', {'flow_classifier': {'name': 'test2'}},
                    fc['flow_classifier']['id']
                )
                updated_flow_classifier = copy.copy(original_flow_classifier)
                if updated:
                    updated_flow_classifier['name'] = 'test2'
                res = req.get_response(self.ext_api)
                self.assertEqual(500, res.status_int)
                driver_manager = self.fake_driver_manager
                (driver_manager.update_flow_classifier_precommit
                 .assert_called_once_with(mock.ANY))
                if updated:
                    (driver_manager.update_flow_classifier_postcommit
                     .assert_called_once_with(mock.ANY))
                else:
                    (driver_manager.update_flow_classifier_postcommit
                     .assert_not_called())
                res = self._list('flow_classifiers')
                self.assertIn('flow_classifiers', res)
                self.assertCountEqual(
                    res['flow_classifiers'], [updated_flow_classifier])

    def test_update_flow_classifier_precommit_driver_manager_exception(self):
        self.fake_driver_manager.update_flow_classifier_precommit = mock.Mock(
            side_effect=fc_exc.FlowClassifierDriverError(
                method='update_flow_classifier_precommit'
            )
        )
        self._test_update_flow_classifier_driver_manager_exception(False)

    def test_update_flow_classifier_postcommit_driver_manager_exception(self):
        self.fake_driver_manager.update_flow_classifier_postcommit = mock.Mock(
            side_effect=fc_exc.FlowClassifierDriverError(
                method='update_flow_classifier_postcommit'
            )
        )
        self._test_update_flow_classifier_driver_manager_exception(True)

    def test_delete_flow_classifer_driver_manager_called(self):
        self.fake_driver_manager.delete_flow_classifier = mock.Mock(
            side_effect=self._record_context)
        self.fake_driver_manager.delete_flow_classifier_precommit = mock.Mock(
            side_effect=self._record_context_precommit)
        self.fake_driver_manager.delete_flow_classifier_postcommit = mock.Mock(
            side_effect=self._record_context_postcommit)
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(
                flow_classifier={'logical_source_port': port['port']['id']},
                do_delete=False
            ) as fc:
                req = self.new_delete_request(
                    'flow_classifiers', fc['flow_classifier']['id']
                )
                res = req.get_response(self.ext_api)
                self.assertEqual(204, res.status_int)
                driver_manager = self.fake_driver_manager
                (driver_manager.delete_flow_classifier
                 .assert_called_once_with(mock.ANY))
                (driver_manager.delete_flow_classifier_precommit
                 .assert_called_once_with(mock.ANY))
                (driver_manager.delete_flow_classifier_postcommit
                 .assert_called_once_with(mock.ANY))
                self.assertIsInstance(
                    self.plugin_context, fc_ctx.FlowClassifierContext
                )
                self.assertIsInstance(
                    self.plugin_context_precommit, fc_ctx.FlowClassifierContext
                )
                self.assertIsInstance(self.plugin_context_postcommit,
                                      fc_ctx.FlowClassifierContext)
                self.assertIn('flow_classifier', fc)
                self.assertEqual(
                    self.plugin_context.current, fc['flow_classifier'])
                self.assertEqual(self.plugin_context_precommit.current,
                                 fc['flow_classifier'])
                self.assertEqual(self.plugin_context_postcommit.current,
                                 fc['flow_classifier'])

    def _test_delete_flow_classifier_driver_manager_exception(self):
        with self.port(
            name='test1'
        ) as port:
            with self.flow_classifier(flow_classifier={
                'name': 'test1',
                'logical_source_port': port['port']['id']
            }, do_delete=False) as fc:
                req = self.new_delete_request(
                    'flow_classifiers', fc['flow_classifier']['id']
                )
                res = req.get_response(self.ext_api)
                self.assertEqual(500, res.status_int)
                driver_manager = self.fake_driver_manager
                driver_manager.delete_flow_classifier.assert_called_once_with(
                    mock.ANY
                )
                self._test_list_resources('flow_classifier', [fc])

    def test_delete_flow_classifier_driver_manager_exception(self):
        self.fake_driver_manager.delete_flow_classifier = mock.Mock(
            side_effect=fc_exc.FlowClassifierDriverError(
                method='delete_flow_classifier'
            )
        )
        self._test_delete_flow_classifier_driver_manager_exception()

    def test_delete_flow_classifier_precommit_driver_manager_exception(self):
        self.fake_driver_manager.delete_flow_classifier_precommit = mock.Mock(
            side_effect=fc_exc.FlowClassifierDriverError(
                method='delete_flow_classifier_precommit'
            )
        )
        self._test_delete_flow_classifier_driver_manager_exception()



# Copyright (c) 2016 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.tests.functional.agent.l2 import base


class TestOVSAgentSfcExtension(base.OVSAgentTestFramework):
    def setUp(self):
        super(TestOVSAgentSfcExtension, self).setUp()
        self.config.set_override('extensions', ['sfc'], 'agent')
        self.agent = self.create_agent()

    def test_run(self):
        self.agent._report_state()
        agent_state = self.agent.state_rpc.report_state.call_args[0][1]
        self.assertEqual(['sfc'], agent_state['configurations']['extensions'])



# Copyright 2015 Futurewei. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Defining OVS data-model

Revision ID: 5a475fc853e6
Revises: c3e178d4a985
Create Date: 2015-09-30 18:00:57.758762

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '5a475fc853e6'
down_revision = 'c3e178d4a985'


def upgrade():
    op.create_table('sfc_portpair_details',
                    sa.Column('tenant_id', sa.String(length=255),
                              nullable=True),
                    sa.Column('id', sa.String(length=36),
                              nullable=False),
                    sa.Column('ingress', sa.String(length=36),
                              nullable=True),
                    sa.Column('egress', sa.String(length=36),
                              nullable=True),
                    sa.Column('host_id', sa.String(length=255),
                              nullable=False),
                    sa.Column('mac_address', sa.String(length=32),
                              nullable=False),
                    sa.Column('network_type', sa.String(length=8),
                              nullable=True),
                    sa.Column('segment_id', sa.Integer(),
                              nullable=True),
                    sa.Column('local_endpoint', sa.String(length=64),
                              nullable=False),
                    sa.PrimaryKeyConstraint('id')
                    )

    op.create_index(
        op.f('ix_sfc_portpair_details_tenant_id'),
        'sfc_portpair_details', ['tenant_id'], unique=False
    )
    op.create_table('sfc_uuid_intid_associations',
                    sa.Column('id', sa.String(length=36), nullable=False),
                    sa.Column('uuid', sa.String(length=36), nullable=False),
                    sa.Column('intid', sa.Integer(), nullable=False),
                    sa.Column('type_', sa.String(length=32), nullable=False),
                    sa.PrimaryKeyConstraint('id', 'uuid'),
                    sa.UniqueConstraint('intid')
                    )

    op.create_table('sfc_path_nodes',
                    sa.Column('tenant_id', sa.String(length=255),
                              nullable=True),
                    sa.Column('id', sa.String(length=36),
                              nullable=False),
                    sa.Column('nsp', sa.Integer(),
                              nullable=False),
                    sa.Column('nsi', sa.Integer(),
                              nullable=False),
                    sa.Column('node_type', sa.String(length=32),
                              nullable=True),
                    sa.Column('portchain_id', sa.String(length=255),
                              nullable=True),
                    sa.Column('status', sa.String(length=32),
                              nullable=True),
                    sa.Column('next_group_id', sa.Integer(),
                              nullable=True),
                    sa.Column('next_hop', sa.String(length=512),
                              nullable=True),
                    sa.ForeignKeyConstraint(['portchain_id'],
                                            ['sfc_port_chains.id'],
                                            ondelete='CASCADE'),
                    sa.PrimaryKeyConstraint('id')
                    )
    op.create_index(
        op.f('ix_sfc_path_nodes_tenant_id'),
        'sfc_path_nodes', ['tenant_id'], unique=False
    )

    op.create_table('sfc_path_port_associations',
                    sa.Column('pathnode_id', sa.String(length=36),
                              nullable=False),
                    sa.Column('portpair_id', sa.String(length=36),
                              nullable=False),
                    sa.Column('weight', sa.Integer(),
                              nullable=False),
                    sa.ForeignKeyConstraint(['pathnode_id'],
                                            ['sfc_path_nodes.id'],
                                            ondelete='CASCADE'),
                    sa.ForeignKeyConstraint(['portpair_id'],
                                            ['sfc_portpair_details.id'],
                                            ondelete='CASCADE'),
                    sa.PrimaryKeyConstraint('pathnode_id', 'portpair_id')
                    )
